package net.novagamestudios.kaffeekasse.ui.util

import androidx.compose.ui.graphics.Color
import androidx.compose.ui.text.AnnotatedString
import androidx.compose.ui.text.SpanStyle
import androidx.compose.ui.text.input.TransformedText

object MacAddress {

    private const val ChunkCount = 6
    private const val ChunkLength = 2
    private const val ChunkSeparator = ":"

    fun String.compress() = this
        .filter { it.isLetterOrDigit() }
        .take(ChunkCount * ChunkLength)
        .uppercase()

    private fun String.chunks() = this
        .compress()
        .chunked(ChunkLength)

    fun String.decompress(): String {
        return chunks().joinToString(ChunkSeparator)
    }

    val String.isValid: Boolean get() {
        return compress().length == ChunkCount * ChunkLength
    }

    class VisualTransformation(
        private val placeholderColor: Color,
        private val placeholder: Char = 'X'
    ) : androidx.compose.ui.text.input.VisualTransformation {
        override fun filter(original: AnnotatedString): TransformedText {
            val chunks = original.text.chunks()
            val transformed = AnnotatedString.Builder().apply {
                for (i in 0 until ChunkCount) {
                    val chunk = chunks.getOrElse(i) { "" }
                    append(chunk)
                    val missing = ChunkLength - chunk.length
                    if (missing > 0) {
                        pushStyle(SpanStyle(color = placeholderColor))
                        append("$placeholder".repeat(missing))
                        pop()
                    }
                    if (i < ChunkCount - 1) {
                        append(ChunkSeparator)
                    }
                }
            }.toAnnotatedString()

            return TransformedText(
                text = transformed,
                offsetMapping = OffsetMapping(original.text)
            )
        }
    }

    private class OffsetMapping(
        private val originalText: String
    ) : androidx.compose.ui.text.input.OffsetMapping {
        override fun originalToTransformed(offset: Int): Int {
            return (offset + offset / ChunkLength).coerceIn(0, (ChunkLength + 1) * ChunkCount - 1)
        }
        override fun transformedToOriginal(offset: Int): Int {
            return (offset - offset / (ChunkLength + 1)).coerceIn(0, originalText.length)
        }
    }
}
